/*
 * loader.c --		Target-side bootstrap loader for CL-PS7111, EP7211.
 *
 * Copyright (C) 1999 Ben Williamson <benw@pobox.com>
 *
 * This program is loaded into SRAM in bootstrap mode, where it waits
 * for commands on UART1 to read and write memory, jump to code etc.
 * A design goal for this program is to be entirely independent of the
 * target board.  Anything with a CL-PS7111 or EP7211 should be able to run
 * this code in bootstrap mode.  All the board specifics can be handled on
 * the host.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ioregs.h"
#include "cs8900.h"

#define DRAM_START	((unsigned *)0xc0000000)
#define DRAM_END	((unsigned *)0xe0000000)
#define STEP_BYTES	(256*1024)
#define STEP_WORDS	(STEP_BYTES / 4)
#define PATTERN		0x12345678

unsigned char * patch_locs[] = {
  0xC000EA45,
  0xC000EA46,
  0xC000678f,
  0xc0005570,
  0xc000558C,
  0xc000559C,
  0xc00055b0,
  0xc000557C,
  0xc00055bc,
  0xc0005dac,
  0xc0005dc0,
  0xc0005e00,
  0xc0005e44,
  0xc0005e50,
  0xc0005e88,
  0xc0006034,
  0xc0006040,
  0xc0006050,
  0xc0006060,
  0xc0006074,
  0xc0006080,
  0xc00060bc,
  0xc00060e0,
  0xc0006260,
  0xc00062b0,
  0xc00069b4,
  0xc00062f8,
  0xc000631c,
  0
};

unsigned char patch_bytes[] = {
  'x','x',0x10,
  0x22,0x32, 0x42, 0x52, 0x62, 0x72, 0xe2, 0xe2, 0x12,
  0x12, 0x72, 0x15,
  0x22,0x32, 0x42, 0x52, 0x62, 0x72, 0x02, 0x12, 0x17, 0x17,
  0x17, 0x10, 0x10, 
  0
};

void do_patches(void) {
  int i;
  for(i=0;patch_locs[i];i++) *patch_locs[i]=patch_bytes[i];
}

extern void flush_v3(void);
extern void flush_v4(void);


void memcpy(unsigned char *dest, unsigned char *src, unsigned int count) {
  int i;
  for(i=0;i<count;i++) dest[i]=src[i];
}


static unsigned char get_char(void)
{
	while (IO_SYSFLG1 & URXFE1);
	return IO_UARTDR1 & 0xff;
}

static unsigned char get_char2(void)
{
	while (IO_SYSFLG2 & URXFE2);
	return IO_UARTDR2 & 0xff;
}

static unsigned char check_char2(void)
{
	return !(IO_SYSFLG2 & URXFE2);
}

static void drain(void)
{
	while (IO_SYSFLG1 & UTXFF1);
}

static void drain2(void)
{
	while (IO_SYSFLG2 & UTXFF2);
}

static void put_char(unsigned char c)
{
	drain();
	IO_UARTDR1 = c;
}

static void put_char2(unsigned char c)
{
	drain2();
	IO_UARTDR2 = c;
}

static unsigned get_word(void)
{
	unsigned w;
	
	w = get_char();
	w |= get_char() << 8;
	w |= get_char() << 16;
	w |= get_char() << 24;
	return w;
}

static void put_word(unsigned w)
{
	put_char(w);
	put_char(w >> 8);
	put_char(w >> 16);
	put_char(w >> 24);
}

static void put_nibble_in_hex (unsigned char n) {
  if(n<10) put_char ('0'+n);
  else put_char ('A'-10+n);
}

static void put_byte_in_hex(unsigned char c) {
  put_nibble_in_hex(c>>4);
  put_nibble_in_hex(c&0xF);
}

void put_addr(unsigned int addr) {
  put_byte_in_hex((addr>>24)&0xff);
  put_byte_in_hex((addr>>16)&0xff);
  put_byte_in_hex((addr>>8)&0xff);
  put_byte_in_hex(addr&0xff);
  put_char(':');
  put_char(' ');
}

void dump_line(unsigned int start, unsigned int length) {
  unsigned char * startptr=start;
  int i,j;
  for(i=0;i<length;) {
    put_addr(start+i);
    for(j=0;j<16;j++) {
      put_byte_in_hex(startptr[i+j]);
      put_char(' ');
    }
    put_char('\n');
    i+=j;
  }
}

/*
 * Received:	start address
 *		length
 *
 * Transmitted:	data bytes
 *		checksum byte (sum of transmitted data)
 */
static void read_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		put_char(*p);
		checksum += *p++;
	}
	put_char(checksum);
}


/*
 * Received:	start address
 *		length
 *		data bytes...
 *
 * Transmitted:	checksum byte (sum of received data)
 */
static void write_block(void)
{
	unsigned char checksum = 0;
	unsigned char *p = (unsigned char*) get_word();
	unsigned length = get_word();

	while (length--) {
		*p = get_char();
		checksum += *p++;
	}
	put_char(checksum);
}

int
main(void)
{
	void (*code)(int r0, int r1, int r2, int r3);
	int r0, r1, r2, r3;
	volatile unsigned *w;
	volatile unsigned char *b;


	put_char('!');
	memcpy(0xC0005000, 0x70000000, 0x40000);
	put_char('1');
	do_patches();
	put_char('2');
	put_char('3');
	put_char('4');
	put_char('5');


	//	dump_line(0xc0005000,500);
	IO_PDDDR &=0xFE;


	IO_PDDR |=1;
	
	IO_SYSCON2 |=0x100;
	
	IO_UBRLCR2 = 0x7001;
	
	//	put_char2(0x70);
	put_char('!');
	code=0xc00062d0;
	code(0,0,0,0);

	while (1) {
		switch (get_char()) {
		case '3':	/* Flush v3 MMU */
			flush_v3();
			break;
			
		case '4':	/* Flush v4 MMU */
			flush_v4();
			break;
			
		case 'a':	/* Ack please */
			put_char('+');	/* we're fine, thanks */
			break;
			
		case 'c':	/* Call (address, r0, r1, r2, r3) */
			code = (void *)get_word();
			r0 = get_word();
			r1 = get_word();
			r2 = get_word();
			r3 = get_word();
			drain();
			code(r0, r1, r2, r3);
			break;
		
		
		case 'g':	/* Get single byte (address) */
			put_char(*(unsigned char *)get_word());
			break;
		
		case 's':	/* Set single byte (address, data) */
			b = (unsigned char *)get_word();
			*b = get_char();
			break;
		
		case 'r':	/* Read single word (address) */
			put_word(*(unsigned *)get_word());
			break;
		
		case 'w':	/* Write single word (address, data) */
			w = (unsigned *)get_word();
			*w = get_word();

			break;
		
		case 'R':	/* Read block */
			read_block();
			break;
		
		case 'W':	/* Write block */
			write_block();
			put_char2(0xce);

			break;

		default:
			put_char('?');
			break;
		}
	}
}

